module Maths
  ( dotP,
    lengthSquared,
    unitVector,
    randV3,
    rand,
  )
where

import Control.Monad.Random
import Control.Monad.State (State, evalState, get, put)
import Debug.Trace
import Linear.V3
import Linear.Vector
import System.Random

rand :: (RandomGen g) => Rand g Double
rand = getRandomR (-1.0, 1.0 :: Double)

randV3 :: (RandomGen g) => Rand g (V3 Double)
randV3 = do
  x <- rand
  y <- rand
  z <- rand
  return $ V3 x y z

--  _ -> randV3

--inf = read "Infinity"

degToRadians :: (Floating a) => a -> a
degToRadians d = (d * pi) / 180.0

dotP :: (Num a) => V3 a -> V3 a -> a
dotP (V3 a b c) (V3 d e f) = a * d + b * e + c * f

lengthSquared :: (Num a) => V3 a -> a
lengthSquared (V3 x y z) = x ^ 2 + y ^ 2 + z ^ 2

vectorLength :: (Floating f) => V3 f -> f
vectorLength v = sqrt (lengthSquared v)

unitVector :: (Floating f) => V3 f -> V3 f
unitVector v = v ^/ vectorLength v
